indexing

	description: "Class that splits an URI into its components."

	standards:
		"1. Based upon RFC 2396.%
		%2. Also takes into account the revised version:%
		%      http://www.apache.org/~fielding/uri/rev-2002/rfc2396bis.html%
		%3. See for issues and other examples:%
		%     http://www.apache.org/~fielding/uri/rev-2002/issues.html"

	author: "Berend de Boer"
	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #1 $"

class

	EPX_URI


inherit

	EPX_EXTERNAL_HELPER

	EXCEPTIONS
		export
			{NONE} all
		end

	KL_IMPORTED_STRING_ROUTINES


creation

	make,
	make_resolve


feature -- Initialization.

	make (a_reference: STRING) is
			-- Create an absolute or relative URI.
		require
			reference_not_empty: a_reference /= Void and then not a_reference.is_empty
		do
			full_reference := a_reference
			parse_reference
		end

	make_resolve (base: EPX_URI; a_reference: STRING) is
			-- If `a_reference' is a partial URI, it is resolved using `base'.
		require
			base_not_void: base /= Void
			base_is_absolute_URI: base.is_absolute
			reference_not_void: a_reference /= Void
		do
			full_reference := a_reference
			parse_reference
			-- See steps in RFC 2396, section 5.2
			-- Step 1:

			if is_relative then
				-- Resolve stuff from `base' we don't have

					check
						scheme_empty: scheme = Void
					end

				-- Step 2:
				if scheme = Void and authority = Void and path = Void and query = Void then
					scheme := base.scheme
					authority := base.authority
					path := base.path

				else
					-- Step 3:
					scheme := base.scheme

					if authority = Void then
						authority := base.authority
						if not has_absolute_path then
							-- To resolve
							if path = Void then
								-- 1. we don't have a path, use path of parent, if any
								path := clone (base.path)
							elseif base.path = Void then
								-- 2. Parent doesn't have a path, make our path absolute
								STRING_.insert_character (path, '/', 1)
									check
										absolute: has_absolute_path
									end
							else
								-- 3. Something to resolve
								resolve_path (base.path)
							end
						end
					end
				end

				-- Build `reference' from parsed components
				full_reference := sh.make_with_capacity (64)
				full_reference.append_string (scheme)
				full_reference.append_character (':')
				if authority /= Void then
					full_reference.append_string ("//")
					full_reference.append_string (authority)
				end
				if path /= Void then
					-- path is void for relative //g
					full_reference.append_string (path)
				end
				if query /= Void then
					full_reference.append_string ("?")
					full_reference.append_string (query)
				end
				if fragment /= Void then
					full_reference.append_string ("#")
					full_reference.append_string (fragment)
				end

			end
		ensure
			must_be_absolute: is_absolute
		end


feature -- Queries

	is_absolute: BOOLEAN is
			-- True if this is an absolute URI.
		do
			Result := scheme /= Void
		ensure
			consistent: Result implies not is_relative
		end

	is_relative: BOOLEAN is
			-- True if this is a relative URI.
		do
			Result := scheme = Void
		ensure
			consistent: Result implies not is_absolute
		end

	has_absolute_path: BOOLEAN is
			-- True if this URI has an absolute path.
		do
			Result :=
				path /= Void and then
				path.item (1) = '/'
		end


feature -- Most generic URI components

	full_reference: STRING
			-- The entire thing.

	scheme: STRING
			-- Scheme used, like "http" or "ftp", anything before the ':'.

	scheme_specific_part: STRING
			-- Interpretation depends on scheme, everything after the ':'
			-- and before the '?'


feature -- If URI has a hierarchical relationships within the namespace

	authority: STRING
			-- Authority part of `scheme_specific_part', usually a host name.
			-- It can be more complex however like: <userinfo>@<host>:<port>.
			-- Use `parse_authority' to split authority in these
			-- components if that is applicable for the protocol.

	path: STRING
			-- Path in `scheme_specific_part', consisting of names
			-- separated by slashes.

	query: STRING
			-- Anything after the '?' if present, else Void

	fragment: STRING
			-- The part after the '#' if present, else Void


feature -- If authority is <userinfo>@<host>:<port>

	user_info: STRING
			-- Usually a user name.

	host: STRING
			-- hostname or IP4 address. IP6 addresses are explicitly not
			-- supported by RFC 2396

	port: INTEGER
			-- TCP port, 0 if no port present.

	is_server_authority: BOOLEAN is
			-- True if authority can be parsed as:
			-- [ userinfo '@' ] host [ ':' port ]
			-- and port, if present, is an integer.
		local
			p, q: INTEGER
			has_host: BOOLEAN
		do
			Result := authority /= Void and then not authority.is_empty
			if Result then
				p := authority.index_of ('@', 1)
				q := authority.index_of (':', p + 1)
				has_host :=
					(q = 0 implies authority.count - p > 0) or else
					(q - p > 0)
				Result := has_host
				if Result and then q /= 0 and then q < authority.count then
					Result := authority.substring (q + 1, authority.count).is_integer
				end
			end
		end

	parse_authority (default_port: INTEGER) is
			-- Assume authority can be parsed as:
			-- [ userinfo '@' ] host [ ':' port ].
			-- If assumption is untrue, you get a nice exception...
			-- `default_port' is 0 means no default.
		require
			valid_default_port: default_port >= 0 and default_port <= 65535
			valid_authority: is_server_authority
		local
			p: INTEGER
			host_port: STRING
			s: STRING
			user_info_present: BOOLEAN
		do
			-- scan for userinfo
			p := authority.index_of ('@', 1)
			user_info_present := p > 1
			if user_info_present then
				user_info := authority.substring (1, p - 1)
			end

			-- host_port is remainder of authority
			if p = 0 then
				host_port := authority
			else
				if p + 1 >= authority.count then
					raise ("No host present in authority %"" + authority + "%".")
				else
					host_port := authority.substring (p + 1, authority.count)
				end
			end
				check
					host_port_set_correctly: host_port /= Void implies host_port.count = authority.count - p
				end

			-- scan for port
			p := host_port.index_of (':', 1)
			if p = 0 then
				host := clone (host_port)
				port := default_port
			elseif p = 1 then
				raise ("No host present in authority %"" + authority + "%".")
			else
				host := host_port.substring (1, p - 1)
				if p = host_port.count then
					port := default_port
				else
					s := host_port.substring (p + 1, host_port.count)
					if s.is_integer then
						port := s.to_integer
					else
						raise ("Port in authority %"" + authority + "%" is not an integer.")
					end
				end
			end
		ensure
			user_info_occurs_in_authority: user_info /= Void implies authority.substring_index (user_info, 1) /= 0
			host_not_void: host /= Void
			host_occurs_in_authority: host /= Void implies authority.substring_index (host, 1) /= 0
		end


feature {NONE} -- URI parsing

	State_scheme,
	State_authority,
	State_path,
	State_query,
	State_fragment: INTEGER is unique
			-- States uses in parsing the URI

	clear_url_parts is
			-- Reset all fields to their default state.
		do
			scheme := Void
			authority := Void
			path := Void
			query := Void
			fragment := Void
			user_info := Void
			host := Void
			port := 0
		end

	parse_reference is
			-- Parse reference according to RFC 2396.
		require
			have_reference: full_reference /= Void
		local
			i: INTEGER
			start: INTEGER
			c: CHARACTER
			state: INTEGER
			something_left_to_parse: BOOLEAN
		do
			clear_url_parts
			from
				start := 1
				state := State_scheme
				i := 1
			variant
				(full_reference.count + 1) - i
			until
				i > full_reference.count
			loop
				c := full_reference.item (i)

				-- switch to next state when char encountered that is
				-- forbidden in state.
				-- If `c' is a valid start character for a state is
				-- checked in the next block
				inspect c
				when ':' then
					if state = State_scheme then
						stop_scheme (start, i)
						start := i + 1
						i := i + 1
						c := full_reference.item (i)
						state := State_authority
					end
				when '/' then
					inspect state
					when State_scheme then
						if i = start then
							state := State_authority
						else
							state := State_path
						end
					when State_authority then
						if i > start + 1 then
							stop_authority (start, i)
							start := i
							state := State_path
						end
					else -- ok
					end
				when '?' then
					inspect state
					when State_scheme then
						stop_path (start, i)
					when State_authority then
						if i = start + 1 then
							stop_path (start, i) -- match for /?
						elseif i > start + 1 then
							stop_authority (start, i)
						end
					when State_path then
						if i > start then
							stop_path (start, i)
						end
					else -- ok
					end
					start := i
					state := State_query
				when '#' then
					inspect state
					when State_scheme then
						stop_path (start, i)
					when State_authority then
						if i = start + 1 then
							stop_path (start, i) -- match for /#
						elseif i > start + 1 then
							stop_authority (start, i)
						end
					when State_path then
						stop_path (start, i)
					when State_query then
						stop_query (start, i)
					else -- ok
					end
					start := i
					state := State_fragment
				else -- ok
				end

				-- check here required start characters for a state
				-- move to next state if mismatch
				if state = State_authority then
					-- check if started with //
					if c /= '/' and i < start + 2 then
						state := State_path
					end
				end

				if state = State_path then
					-- check if started with /
					if i = start and (c = '?' or c = '#') then
						state := State_query
					end
				end

				if state = State_query then
					-- check if started with ?
					if c /= '?' and i = start then
						state := State_fragment
					end
				end

				if state = State_fragment then
					-- check if started with #
					if c /= '#' and i = start then
						-- what?? probably not parseable
					end
				end

				i := i + 1
			end

			-- handle last part of string
			something_left_to_parse := start < i
			if something_left_to_parse then
				inspect state
				when State_scheme then
					stop_path (start, i)
				when State_authority then
					if i = start + 1 then
						stop_path (start, i) -- got only a '/'
					elseif i > start + 1 then
						stop_authority (start, i)
					end
				when State_path then
					stop_path (start, i)
				when State_query then
					stop_query (start, i)
				when State_fragment then
					stop_fragment (start, i)
				end
			end

		end

	stop_scheme (start, stop: INTEGER) is
			-- Called when scheme part parsed.
			-- Start is inclusive, stop is exclusive.
		require
			valid_start: start = 1
			valid_stop: stop >= 1 and stop < full_reference.count
		do
			if stop >= start then
				scheme := full_reference.substring (start, stop-1)
			end
			scheme_specific_part := full_reference.substring (stop, full_reference.count)
		end

	stop_authority (start, stop: INTEGER) is
			-- Start is inclusive, stop is exclusive.
		require
			valid_start: full_reference.valid_index (start)
			valid_authority:
				full_reference.item (start) = '/' and
				full_reference.item (start + 1) = '/'
			valid_stop: full_reference.valid_index (stop - 1)
			full_reference_contains_authority: start + 2 <= stop - 1
		do
			authority := full_reference.substring (start+2, stop-1)
		ensure
			authority_set: authority /= Void and then not authority.is_empty
		end

	stop_path (start, stop: INTEGER) is
			-- Start is inclusive, stop is exclusive.
		require
			valid_start: full_reference.valid_index (start)
			valid_stop: stop > start implies full_reference.valid_index (stop-1)
		do
			if stop > start then
				path := full_reference.substring (start, stop-1)
			end
		ensure
			conditional_path_set: stop > start implies path /= Void and then not path.is_empty
			conditional_path_not_set: stop <= start implies path = Void
		end

	stop_query (start, stop: INTEGER) is
			-- Start is inclusive, stop is exclusive.
		require
			valid_start: full_reference.valid_index (start)
			valid_query: full_reference.item (start) = '?'
			valid_stop: full_reference.valid_index (stop - 1)
			full_reference_contains_query: start + 1 <= stop - 1
		do
			query := full_reference.substring (start + 1, stop-1)
		ensure
			query_set: query /= Void and then not query.is_empty
		end

	stop_fragment (start, stop: INTEGER) is
			-- Start is inclusive, stop is exclusive.
		require
			valid_start: full_reference.valid_index (start)
			valid_fragment: full_reference.item (start) = '#'
			valid_stop: full_reference.valid_index (stop - 1)
			full_reference_contains_fragment: start + 1 <= stop - 1
		do
			fragment := full_reference.substring (start + 1, stop-1)
		ensure
			fragment_set: fragment /= Void and then not fragment.is_empty
		end


feature {NONE} -- Resolve a relative-path reference

	resolve_path (base: STRING) is
			-- Resolve path according to rfc2396bis. Abnormal cases are
			-- compensated for.
		require
			base_path_not_void: base /= Void
			something_to_resolve: path /= Void
			-- `path' is relative
		local
			resolved_path: STRING
			i: INTEGER
			base_paths: ARRAY [STRING]
			relative_paths: ARRAY [STRING]
			segment: STRING
		do
			-- see rfc2396, section 5.2 step 6 for an implementation
			-- we use a different one

			base_paths := sh.split_on (base, '/')
			relative_paths := sh.split_on (path, '/')

			-- All but the last segment of the base URI's path component is
			-- copied to the buffer.  In other words, any characters after the
			-- last (right-most) slash character, if any, are excluded.
			if base_paths.upper >= 0 then
				base_paths.resize (base_paths.lower, base_paths.upper - 1)
			end

			from
				i := relative_paths.lower
			until
				i > relative_paths.upper
			loop
				segment := relative_paths.item (i)
				if segment.count = 0 or else segment.is_equal (".") then
					-- do nothing
				elseif segment.is_equal ("..") then
					-- back one level
					if base_paths.count > 0 then
						base_paths.resize (base_paths.lower, base_paths.upper - 1)
					end
				else
					-- add segment
					base_paths.force (segment, base_paths.upper + 1)
				end
				i := i + 1
			end

			-- build path, skip empty segments (just /'s)
			resolved_path := sh.make_with_capacity (base.count)
			from
				i := base_paths.lower
			until
				i > base_paths.upper
			loop
				segment := base_paths.item (i)
				if not segment.is_empty then
					resolved_path.append_string ("/")
					resolved_path.append_string (base_paths.item (i))
				end
				i := i + 1
			end

			-- make sure we have a / at the end if needed

			if relative_paths.item (relative_paths.count - 1).is_empty or else
				relative_paths.item (relative_paths.count - 1).is_equal (".") or else
				relative_paths.item (relative_paths.count - 1).is_equal ("..") then
				resolved_path.append_character ('/')
			end

			path := resolved_path
		end

invariant

	either_absolute_or_relative: is_absolute xor is_relative
	reference_not_empty: full_reference /= Void and then not full_reference.is_empty

-- I'm really unsure if these constraints hold for deliberate garbage...

	-- Constraints on elements of a parsed URI.
	valid_authority: authority = Void or else not authority.is_empty
	valid_path: path = Void or else not path.is_empty
	valid_query: query = Void or else not query.is_empty
	valid_fragment: fragment = Void or else not fragment.is_empty

	-- Contraints on parsed `authority'
	user_info_occurs_in_authority: user_info /= Void implies authority.substring_index (user_info, 1) /= 0
	host_occurs_in_authority: host /= Void implies authority.substring_index (host, 1) /= 0
	valid_port: port >= 0 and port <= 65535

end
