indexing

	description: "Lexical analyzer input file buffers for STDC_FILE."

	author: "Berend de Boer"
	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #1 $"

class

	EPX_FILE_BUFFER

inherit

	YY_BUFFER
		rename
			make as make_from_string
		redefine
			content,
			fill,
			name,
			new_default_buffer,
			wipe_out
		end

creation

	make, make_with_size

feature {NONE} -- Initialization

	make (a_file: like file) is
			-- Create a new buffer for `a_file'.
		require
			a_file_not_void: a_file /= Void
			a_file_open: a_file.is_open
		do
			make_with_size (a_file, Read_buffer_capacity)
		ensure
			count_set: count = 0
			file_set: file = a_file
			beginning_of_line: beginning_of_line
		end

	make_with_size (a_file: like file; size: INTEGER) is
			-- Create a new buffer of capacity `size' for `a_file'.
		require
			a_file_not_void: a_file /= Void
			a_file_open: a_file.is_open
			size_positive: size >= 0
			size_large_enough: size >= 512
		do
			capacity := size
				-- `content' has to be 2 characters longer
				-- than the size given because we need to
				-- put in 2 end-of-buffer characters.
			create content.make (size + 2)
			set_file (a_file)
		ensure
			capacity_set: capacity = size
			count_set: count = 0
			file_set: file = a_file
			beginning_of_line: beginning_of_line
		end

feature -- Access

	name: STRING is
			-- Name of buffer.
		do
			Result := file.filename
		end

	content: EPX_CHARACTER_BUFFER
			-- Input buffer characters.

	file: STDC_TEXT_FILE
			-- Input file. Should have buffering disabled.

feature -- Status report

	end_of_file: BOOLEAN
			-- Has end-of-file been reached?

feature -- Setting

	set_file (a_file: like file) is
			-- Set `file' to `a_file'.
		require
			a_file_not_void: a_file /= Void
			a_file_open: a_file.is_open
		do
			end_of_file := False
			flush
			file := a_file
		ensure
			count_set: count = 0
			file_set: file = a_file
			beginning_of_line: beginning_of_line
		end

feature -- Element change

	fill is
			-- Fill buffer with characters from `file'.
			-- Do not lose unprocessed characters in buffer.
			-- Resize buffer if necessary. Set `filled' to True if
			-- characters have been added to buffer.
		local
			buff: like content
			bytes: INTEGER
		do
			-- If the last call to `fill' failed to add more characters,
			-- this means that the end of file has already been
			-- reached. Do not attempt to fill again the buffer in that
			-- case.
			if filled and not end_of_file then
				-- First move last characters to start of buffer
				-- and eventually resize `content' if necessary.
				compact_left
				buff := content
				-- Read in more data.
				-- It's best to read in blocks divisible by 512 bytes
				bytes := capacity - count
				bytes := (bytes // 512) * 512
					check
						bytes_positive: bytes >= 0
					end
				file.read_buffer (buff, count, bytes)
				if not file.eof then
					count := count + file.last_read
					filled := True
				else
					filled := False
					end_of_file := True
				end
				buff.put (End_of_buffer_character, count + 1)
				buff.put (End_of_buffer_character, count + 2)
			else
				filled := False
			end
		end

	wipe_out is
			-- Wipe out buffer.
		do
			flush
			filled := False
			end_of_file := True
		ensure then
			end_of_file: end_of_file
		end

feature {NONE} -- Implementation

	new_default_buffer (a_number_of_bytes: INTEGER): like content is
			-- Create a new buffer that is like `content'.
		do
			create Result.make (a_number_of_bytes)
		end

feature {NONE} -- Constants

	Read_buffer_capacity: INTEGER is 8192
			-- Maximum number of characters to be read at a time.

invariant

	file_not_void: file /= Void
	file_open_read: not end_of_file implies file.is_open
	default_capacity_large_enough: Read_buffer_capacity >= 512

end
